import unittest
from unittest.mock import MagicMock, AsyncMock, patch
import json
import asyncio
import sys
import os

# Ensure we can import modules from the src and parent directory
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'src')))
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))

from sec_gemini_byot import byot, hub_pb2

class TestByot(unittest.IsolatedAsyncioTestCase):
    async def test_handle_list_tools(self):
        # Setup mock MCP client
        mock_client = MagicMock()
        
        # Mock tool structure - FastMCP tools usually have a model_dump method
        tool1 = MagicMock()
        tool1.model_dump.return_value = {"name": "tool1", "description": "test tool 1"}
        
        tool2 = MagicMock()
        tool2.model_dump.return_value = {"name": "tool2", "description": "test tool 2"}
        
        mock_client.list_tools = AsyncMock(return_value=[tool1, tool2])
        
        mcp_clients = [mock_client]
        
        # execution
        result_json = await byot._handle_list_tools(mcp_clients)
        
        # verification
        result = json.loads(result_json)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0]["name"], "tool1")
        self.assertEqual(result[1]["name"], "tool2")

    async def test_handle_call_tool_success(self):
        mock_client = MagicMock()
        tool = MagicMock()
        tool.name = "my_tool"
        tool.model_dump.return_value = {"name": "my_tool"}
        mock_client.list_tools = AsyncMock(return_value=[tool])
        
        # Mock result
        mock_result = MagicMock()
        mock_result.model_dump.return_value = {"content": [{"type": "text", "text": "result"}]}
        # Simulate result not having dict method to hit the first if branch
        del mock_result.dict 
        
        mock_client.call_tool = AsyncMock(return_value=mock_result)
        
        mcp_clients = [mock_client]
        
        job = MagicMock()
        job.request_data = json.dumps({
            "name": "my_tool",
            "arguments": {"arg1": "val1"}
        })
        
        result_json = await byot._handle_call_tool(job, mcp_clients)
        
        result = json.loads(result_json)
        self.assertEqual(result["content"][0]["text"], "result")
        mock_client.call_tool.assert_called_with("my_tool", {"arg1": "val1"})

    async def test_handle_call_tool_not_found(self):
        mock_client = MagicMock()
        mock_client.list_tools = AsyncMock(return_value=[])
        mcp_clients = [mock_client]
        
        job = MagicMock()
        job.request_data = json.dumps({
            "name": "missing_tool",
            "arguments": {}
        })
        
        result = await byot._handle_call_tool(job, mcp_clients)
        self.assertIn("Error: Tool missing_tool not found", result)

    async def test_process_jobs_list_tools(self):
        mock_stream = MagicMock()
        # Mock async iterator for stream
        
        job_request = MagicMock()
        job_request.job_id = "job1"
        job_request.request_type = hub_pb2.RequestType.LIST_TOOLS
        
        msg = MagicMock()
        msg.job_request = job_request
        
        # Setup async iterator to return one message then stop
        m = AsyncMock()
        m.__aiter__.return_value = [msg]
        mock_stream = m
        
        mock_client = MagicMock()
        mock_client.list_tools = AsyncMock(return_value=[])
        mcp_clients = [mock_client]
        
        mock_stream.write = AsyncMock()
        
        await byot._process_jobs(mock_stream, mcp_clients)
        
        mock_stream.write.assert_called_once()
        call_args = mock_stream.write.call_args[0][0]
        self.assertEqual(call_args.job_result.job_id, "job1")
        self.assertEqual(call_args.job_result.content, "[]")

    async def test_process_jobs_ping(self):
        job_request = MagicMock()
        job_request.job_id = "job_ping"
        job_request.request_type = hub_pb2.RequestType.PING
        
        msg = MagicMock()
        msg.job_request = job_request
        
        mock_stream = AsyncMock()
        mock_stream.__aiter__.return_value = [msg]
        mock_stream.write = AsyncMock()

        await byot._process_jobs(mock_stream, [])
        
        mock_stream.write.assert_called_once()
        call_args = mock_stream.write.call_args[0][0]
        self.assertEqual(call_args.job_result.content, "pong")

    async def test_process_jobs_error(self):
        job_request = MagicMock()
        job_request.job_id = "job_error"
        job_request.request_type = hub_pb2.RequestType.CALL_TOOL
        # request_data missing will cause json.loads to fail (if we don't mock it)
        # But wait, job is a mock, job.request_data is a mock. json.loads(mock) will fail?
        # json.loads expects string, bytes or bytearray.
        # Let's make it fail by having job.request_data be invalid json string
        job_request.request_data = "{invalid_json"
        
        msg = MagicMock()
        msg.job_request = job_request
        
        mock_stream = AsyncMock()
        mock_stream.__aiter__.return_value = [msg]
        mock_stream.write = AsyncMock()

        await byot._process_jobs(mock_stream, [])
        
        mock_stream.write.assert_called_once()
        call_args = mock_stream.write.call_args[0][0]
        self.assertIn("Error executing tool", call_args.job_result.content)

    @patch("grpc.aio.secure_channel")
    @patch("grpc.aio.insecure_channel")
    @patch("grpc.ssl_channel_credentials")
    @patch("fastmcp.Client")
    async def test_run_client_secure(self, mock_fastmcp, mock_ssl_creds, mock_insecure, mock_secure):
        # Setup mocks
        mock_secure.return_value.__aenter__.return_value = MagicMock()
        mock_fastmcp.return_value.__aenter__.return_value = MagicMock()
        
        # We need to mock the stub and the stream to avoid actual network calls or further errors
        with patch("sec_gemini_byot.hub_pb2_grpc.BringYourOwnToolsHubStub") as mock_stub:
            mock_stream = AsyncMock()
            mock_stream.__aiter__.return_value = [] # Empty stream
            mock_stub.return_value.Connect.return_value = mock_stream
            
            await byot.run_client("key", "name", "https://hub.com", [])
            
        mock_secure.assert_called_once_with("hub.com", mock_ssl_creds.return_value)
        mock_insecure.assert_not_called()

    @patch("grpc.aio.secure_channel")
    @patch("grpc.aio.insecure_channel")
    @patch("fastmcp.Client")
    async def test_run_client_insecure_http(self, mock_fastmcp, mock_insecure, mock_secure):
        mock_insecure.return_value.__aenter__.return_value = MagicMock()
        mock_fastmcp.return_value.__aenter__.return_value = MagicMock()
        
        with patch("sec_gemini_byot.hub_pb2_grpc.BringYourOwnToolsHubStub") as mock_stub:
            mock_stream = AsyncMock()
            mock_stream.__aiter__.return_value = []
            mock_stub.return_value.Connect.return_value = mock_stream
            
            await byot.run_client("key", "name", "http://localhost:50051", [])
            
        mock_insecure.assert_called_once_with("localhost:50051")
        mock_secure.assert_not_called()

    @patch("grpc.aio.secure_channel")
    @patch("grpc.aio.insecure_channel")
    @patch("fastmcp.Client")
    async def test_run_client_insecure_default(self, mock_fastmcp, mock_insecure, mock_secure):
        mock_insecure.return_value.__aenter__.return_value = MagicMock()
        mock_fastmcp.return_value.__aenter__.return_value = MagicMock()
        
        with patch("sec_gemini_byot.hub_pb2_grpc.BringYourOwnToolsHubStub") as mock_stub:
            mock_stream = AsyncMock()
            mock_stream.__aiter__.return_value = []
            mock_stub.return_value.Connect.return_value = mock_stream
            
            await byot.run_client("key", "name", "localhost:50051", [])
            
        mock_insecure.assert_called_once_with("localhost:50051")
        mock_secure.assert_not_called()

if __name__ == '__main__':
    unittest.main()



import asyncio
import logging
import argparse
import socket
from rich.logging import RichHandler
import grpc
import tenacity
try:
    from . import byot
    from . import default_mcp
except ImportError:
    import byot
    import default_mcp

# Constants.
DEFAULT_HUB_URL = "https://bringyourowntool-hub-171354917004.us-central1.run.app"

# Set up logging.
logging.getLogger("grpc").setLevel(logging.WARNING) 
logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True)]
)

# Tenacity here deals with reconnecting to the hub if it goes down temporarely.
@tenacity.retry(
    retry=tenacity.retry_if_exception_type((grpc.RpcError, OSError)), 
    wait=tenacity.wait_exponential(multiplier=1, min=4, max=10)
)
async def main():
    # Parse arguments.
    parser = argparse.ArgumentParser(description="Sec-Gemini Bring-Your-Own-Tool Client")
    parser.add_argument("--name", default=socket.gethostname(), help="Name of this client (default: hostname)")
    parser.add_argument("--api-key", help="Sec-GeminiAPI Key for authentication", required=True)
    parser.add_argument("--hub", default=DEFAULT_HUB_URL, help="Hub address (host:port)")
    parser.add_argument("--use-default-mcp", default=True, action="store_true", help="Use the default MCP")
    parser.add_argument("--additional-mcps", default=[], nargs="+", help="Additional MCPS as urls or file paths")
    args = parser.parse_args()
    # Compile the list of mcps to use.
    mcps = [*([default_mcp.mcp] if args.use_default_mcp else []), *args.additional_mcps]
    # And run the client.
    logging.info("Starting client...")
    if args.use_default_mcp:
        logging.info("Using default MCP")
    try:
        await byot.run_client(args.api_key, args.name, args.hub, mcps)
    except KeyboardInterrupt:
        logging.info("Keyboard interrupt received, I'm done.")

def run():
    asyncio.run(main())

if __name__ == "__main__":
    run()



import asyncio
import contextlib
import grpc
import logging
try:
    from . import hub_pb2
    from . import hub_pb2_grpc
except (ImportError, ValueError):
    import hub_pb2
    import hub_pb2_grpc
import json
import fastmcp


# Set up logging.
logger = logging.getLogger(__name__)

async def _handle_list_tools(mcp_clients) -> str:
    all_tools = []
    for mcp_client in mcp_clients:
        client_tools = await mcp_client.list_tools()
        all_tools.extend([t.model_dump() for t in client_tools])
    return json.dumps(all_tools)

async def _handle_call_tool(job, mcp_clients) -> str:
    request_dict = json.loads(job.request_data)
    logger.info(f"Request Data: {request_dict}")
    tool_name = request_dict.get("name")
    tool_args = request_dict.get("arguments", {})
    
    for mcp_client in mcp_clients:
        client_tools = await mcp_client.list_tools()
        if any(t.name == tool_name for t in client_tools):
            result = await mcp_client.call_tool(tool_name, tool_args)
            # Serialize result robustly handles Pydantic v1/v2 or plain objects
            if hasattr(result, "model_dump"):
                r_dict = result.model_dump()
            elif hasattr(result, "dict"):
                r_dict = result.dict()
            else:
                # Fallback for plain objects
                r_dict = {
                    "content": getattr(result, "content", []),
                    "isError": getattr(result, "isError", False),
                    "_meta": getattr(result, "meta", None),
                    "structuredContent": getattr(result, "structuredContent", getattr(result, "structured_content", None))
                }
                
            # Ensure r_dict is json serializable
            return json.dumps(r_dict, default=str)
    
    return f"Error: Tool {tool_name} not found."

async def _process_jobs(stream, mcp_clients):
    # Listen for jobs.
    async for msg in stream:
        if msg.job_request.job_id:
            job = msg.job_request
            logger.info(f"Received Job: {hub_pb2.RequestType.Name(job.request_type)}")
            
            # Execute Tool via FastMCP logic
            result_content = ""
            try:

                match (job.request_type):
                    case hub_pb2.RequestType.LIST_TOOLS:
                        result_content = await _handle_list_tools(mcp_clients)
                    case hub_pb2.RequestType.CALL_TOOL:
                        result_content = await _handle_call_tool(job, mcp_clients)
                    case hub_pb2.RequestType.PING:
                        result_content = "pong"
                    case _:
                        result_content = "Error: Unknown request type."
            except Exception as e:
                result_content = f"Error executing tool: {e}"
                
            # Send Result
            result_msg = hub_pb2.ClientMessage(
                job_result=hub_pb2.JobResult(
                    job_id=job.job_id,
                    content=str(result_content)
                )
            )
            await stream.write(result_msg)
            logger.info(f"[+] Result sent for {job.job_id}")

        elif msg.pong.msg:
            logger.info(f"[.] Pong: {msg.pong.msg}")

async def run_client(
    api_key: str,
    client_name: str,
    hub_url: str,
    mcp_servers: list[fastmcp.FastMCP | str]):

    async with contextlib.AsyncExitStack() as stack:
        # Create clients with all MCP servers.
        mcp_clients = []
        for server in mcp_servers:
            client = fastmcp.Client(server)
            await stack.enter_async_context(client)
            mcp_clients.append(client)
        # Connect to the hub.
        if hub_url.startswith("https://"):
            target = hub_url[len("https://"):]
            channel = await stack.enter_async_context(
                grpc.aio.secure_channel(target, grpc.ssl_channel_credentials())
            )
        else:
            target = hub_url
            if hub_url.startswith("http://"):
                target = hub_url[len("http://"):]
            channel = await stack.enter_async_context(
                grpc.aio.insecure_channel(target)
            )
        stub = hub_pb2_grpc.BringYourOwnToolsHubStub(channel)
        logger.info(f"Connecting to {hub_url}...")
        try:
            stream = stub.Connect()
            # Send register message.
            logger.info(f"Registering as '{client_name}' with key '{api_key}'...")
            await stream.write(hub_pb2.ClientMessage(
                register=hub_pb2.Register(
                    api_key=api_key,
                    name=client_name
                )
            ))
            logger.info(f"Ready!")
            # Listen for jobs.
            await _process_jobs(stream, mcp_clients)

        except grpc.RpcError as e:
            logger.error(f"Connection error!")
            logger.error(f"gRPC error: {e}")
            raise e
        except asyncio.CancelledError as e:
            logger.error(f"gRPC error: {e}")
            pass



import os
import subprocess
import sys
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py

class BuildPyCommand(build_py):
    def run(self):
        # Path to the proto file
        proto_path = os.path.join("common", "hub.proto")
        
        if os.path.exists(proto_path):
            print(f"Compiling {proto_path}...")
            try:
                import grpc_tools.protoc
                # We want the output to be inside the package
                output_dir = os.path.join("src", "sec_gemini_byot")
                os.makedirs(output_dir, exist_ok=True)
                
                grpc_tools.protoc.main([
                    'grpc_tools.protoc',
                    '-Icommon',
                    f'--python_out={output_dir}',
                    f'--grpc_python_out={output_dir}',
                    proto_path,
                ])
                
                # Fix imports in generated files for Python 3 relative imports
                # The generated hub_pb2_grpc.py might have "import hub_pb2"
                # which needs to be "from . import hub_pb2" if we want it to work as a package.
                generated_grpc = os.path.join(output_dir, "hub_pb2_grpc.py")
                if os.path.exists(generated_grpc):
                    with open(generated_grpc, 'r') as f:
                        content = f.read()
                    content = content.replace('import hub_pb2 as hub__pb2', 'from . import hub_pb2 as hub__pb2')
                    with open(generated_grpc, 'w') as f:
                        f.write(content)
                        
            except ImportError:
                print("grpcio-tools not found during build, skipping proto compilation")
        else:
            print(f"Warning: {proto_path} not found")
        
        super().run()

if __name__ == "__main__":
    setup(
        cmdclass={
            'build_py': BuildPyCommand,
        },
    )



"""Default MCP for the BYOT client.

This allows basic file operations and shell commands."""

from fastmcp import FastMCP
from fastmcp.tools.tool import ToolAnnotations
from typing import Literal
import requests
import subprocess
import pathlib

mcp = FastMCP("BringYourOwnToolMcp")

@mcp.tool(task=True, annotations=ToolAnnotations(destructiveHint=True))
async def shell_async(command: str) -> dict[str, str | bool]:
    result = subprocess.run(command, shell=True, capture_output=True, text=True)
    return {
        "stdout": result.stdout,
        "is_error": result.returncode != 0,
        "stderr": result.stderr
    }

@mcp.tool(annotations={"destructiveHint": False, "readOnlyHint": True})
def list_files(path: str) -> list[str]:
    return [str(p) for p in pathlib.Path(path).iterdir()][:100]

@mcp.tool(annotations={"destructiveHint": False, "readOnlyHint": True})
def read_file(path: str) -> str:
    return pathlib.Path(path).read_text()

@mcp.tool(annotations={"destructiveHint": True, "readOnlyHint": False})
def write_file(path: str, content: str) -> None:
    pathlib.Path(path).write_text(content)

@mcp.tool(annotations={"destructiveHint": True, "readOnlyHint": False})
def remove_file(path: str) -> None:
    pathlib.Path(path).unlink()

@mcp.tool(annotations={"destructiveHint": True, "readOnlyHint": False})
async def remove_directory(path: str) -> None:
    pathlib.Path(path).rmdir()

@mcp.tool(annotations={"destructiveHint": True, "readOnlyHint": False, "openWorldHint": True})
async def fetch(method: Literal['GET', 'POST', 'PUT', 'DELETE', 'PATCH'], url: str) -> bytes:
    response = requests.request(method, url)
    return {
        "content": response.content,
        "status_code": response.status_code
    }